@@ -136,11 +136,11 @@ static void setup_unref(struct a2dp_setup *setup)
 
 static struct audio_device *a2dp_get_dev(struct avdtp *session)
 {
-	bdaddr_t addr;
+	bdaddr_t src, dst;
 
-	avdtp_get_peers(session, NULL, &addr);
+	avdtp_get_peers(session, &src, &dst);
 
-	return manager_find_device(&addr, NULL, FALSE);
+	return manager_find_device(NULL, &src, &dst, NULL, FALSE);
 }
 
 static gboolean finalize_config(struct a2dp_setup *s)
@@ -150,13 +150,15 @@ static gboolean finalize_config(struct a2dp_setup *s)
 	setup_ref(s);
 	for (l = s->cb; l != NULL; l = l->next) {
 		struct a2dp_setup_cb *cb = l->data;
+		struct avdtp_stream *stream = s->err ? NULL : s->stream;
 
-		if (cb->config_cb) {
-			cb->config_cb(s->session, s->sep, s->stream, s->err,
-					cb->user_data);
-			cb->config_cb = NULL;
-			setup_unref(s);
-		}
+		if (!cb->config_cb)
+			continue;
+
+		cb->config_cb(s->session, s->sep, stream, s->err,
+							cb->user_data);
+		cb->config_cb = NULL;
+		setup_unref(s);
 	}
 
 	setup_unref(s);
@@ -181,7 +183,7 @@ static gboolean finalize_resume(struct a2dp_setup *s)
 	for (l = s->cb; l != NULL; l = l->next) {
 		struct a2dp_setup_cb *cb = l->data;
 
-		if (cb->resume_cb) {
+		if (cb && cb->resume_cb) {
 			cb->resume_cb(s->session, s->err, cb->user_data);
 			cb->resume_cb = NULL;
 			setup_unref(s);
@@ -1236,14 +1238,48 @@ gboolean a2dp_source_cancel(struct audio_device *dev, unsigned int id)
 	return TRUE;
 }
 
-unsigned int a2dp_source_config(struct avdtp *session, a2dp_config_cb_t cb,
-				GSList *caps, void *user_data)
+struct a2dp_sep *a2dp_source_get(struct avdtp *session,
+				struct avdtp_remote_sep *rsep)
+{
+	GSList *l;
+	struct a2dp_server *server;
+	struct avdtp_service_capability *cap;
+	struct avdtp_media_codec_capability *codec_cap = NULL;
+	bdaddr_t src;
+
+	avdtp_get_peers(session, &src, NULL);
+	server = find_server(servers, &src);
+	if (!server)
+		return NULL;
+
+	cap = avdtp_get_codec(rsep);
+	codec_cap = (void *) cap->data;
+
+	for (l = server->sources; l != NULL; l = l->next) {
+		struct a2dp_sep *sep = l->data;
+
+		if (sep->locked)
+			continue;
+
+		if (sep->codec != codec_cap->media_codec_type)
+			continue;
+
+		if (!sep->stream || avdtp_has_stream(session, sep->stream))
+			return sep;
+	}
+
+	return NULL;
+}
+
+unsigned int a2dp_source_config(struct avdtp *session, struct a2dp_sep *sep,
+				a2dp_config_cb_t cb, GSList *caps,
+				void *user_data)
 {
 	struct a2dp_setup_cb *cb_data;
 	GSList *l;
 	struct a2dp_server *server;
 	struct a2dp_setup *setup;
-	struct a2dp_sep *sep = NULL, *tmp;
+	struct a2dp_sep *tmp;
 	struct avdtp_local_sep *lsep;
 	struct avdtp_remote_sep *rsep;
 	struct avdtp_service_capability *cap;
@@ -1269,25 +1305,8 @@ unsigned int a2dp_source_config(struct avdtp *session, a2dp_config_cb_t cb,
 	if (!codec_cap)
 		return 0;
 
-	for (l = server->sources; l != NULL; l = l->next) {
-		tmp = l->data;
-
-		if (tmp->locked)
-			continue;
-
-		if (tmp->codec != codec_cap->media_codec_type)
-			continue;
-
-		if (!tmp->stream || avdtp_has_stream(session, tmp->stream)) {
-			sep = tmp;
-			break;
-		}
-	}
-
-	if (!sep) {
-		error("a2dp_source_cfg: no available SEP found");
+	if (sep->codec != codec_cap->media_codec_type)
 		return 0;
-	}
 
 	debug("a2dp_source_config: selected SEP %p", sep->sep);
 
@@ -1345,9 +1364,10 @@ unsigned int a2dp_source_config(struct avdtp *session, a2dp_config_cb_t cb,
 		break;
 	case AVDTP_STATE_OPEN:
 	case AVDTP_STATE_STREAMING:
-		if (avdtp_stream_has_capabilities(setup->stream, caps))
+		if (avdtp_stream_has_capabilities(setup->stream, caps)) {
+			debug("Configuration match: resuming");
 			g_idle_add((GSourceFunc) finalize_config, setup);
-		else if (!setup->reconfigure) {
+		} else if (!setup->reconfigure) {
 			setup->reconfigure = TRUE;
 			if (avdtp_close(session, sep->stream) < 0) {
 				error("avdtp_close failed");
@@ -1515,3 +1535,45 @@ gboolean a2dp_sep_unlock(struct a2dp_sep *sep, struct avdtp *session)
 	return TRUE;
 }
 
+gboolean a2dp_sep_get_lock(struct a2dp_sep *sep)
+{
+	return sep->locked;
+}
+
+static int stream_cmp(gconstpointer data, gconstpointer user_data)
+{
+	const struct a2dp_sep *sep = data;
+	const struct avdtp_stream *stream = user_data;
+
+	return (sep->stream != stream);
+}
+
+struct a2dp_sep *a2dp_get_sep(struct avdtp *session,
+				struct avdtp_stream *stream)
+{
+	struct a2dp_server *server;
+	bdaddr_t src, dst;
+	GSList *l;
+
+	avdtp_get_peers(session, &src, &dst);
+
+	for (l = servers; l; l = l->next) {
+		server = l->data;
+
+		if (bacmp(&src, &server->src) == 0)
+			break;
+	}
+
+	if (!l)
+		return NULL;
+
+	l = g_slist_find_custom(server->sources, stream, stream_cmp);
+	if (l)
+		return l->data;
+
+	l = g_slist_find_custom(server->sinks, stream, stream_cmp);
+	if (l)
+		return l->data;
+
+	return NULL;
+}
