@@ -58,6 +58,7 @@
 #include "telephony.h"
 #include "headset.h"
 #include "glib-helper.h"
+#include "btio.h"
 #include "dbus-common.h"
 #include "../src/adapter.h"
 #include "../src/device.h"
@@ -105,6 +106,12 @@ static char *str_state[] = {
 	"HEADSET_STATE_PLAYING",
 };
 
+struct headset_state_callback {
+	headset_state_cb cb;
+	void *user_data;
+	unsigned int id;
+};
+
 struct connect_cb {
 	unsigned int id;
 	headset_stream_cb_t cb;
@@ -163,6 +170,8 @@ struct event {
 	int (*callback) (struct audio_device *device, const char *buf);
 };
 
+static GSList *headset_callbacks = NULL;
+
 static inline DBusMessage *invalid_args(DBusMessage *msg)
 {
 	return g_dbus_create_error(msg, ERROR_INTERFACE ".InvalidArguments",
@@ -261,6 +270,23 @@ static void print_hf_features(uint32_t features)
 	g_free(str);
 }
 
+static const char *state2str(headset_state_t state)
+{
+	switch (state) {
+	case HEADSET_STATE_DISCONNECTED:
+		return "disconnected";
+	case HEADSET_STATE_CONNECT_IN_PROGRESS:
+		return "connecting";
+	case HEADSET_STATE_CONNECTED:
+	case HEADSET_STATE_PLAY_IN_PROGRESS:
+		return "connected";
+	case HEADSET_STATE_PLAYING:
+		return "playing";
+	}
+
+	return NULL;
+}
+
 static int headset_send_valist(struct headset *hs, char *format, va_list ap)
 {
 	char rsp[BUF_SIZE];
@@ -403,13 +429,16 @@ static void pending_connect_finalize(struct audio_device *dev)
 	struct headset *hs = dev->headset;
 	struct pending_connect *p = hs->pending;
 
+	if (p == NULL)
+		return;
+
 	g_slist_foreach(p->callbacks, (GFunc) pending_connect_complete, dev);
 
 	g_slist_foreach(p->callbacks, (GFunc) g_free, NULL);
 	g_slist_free(p->callbacks);
 
 	if (p->io) {
-		g_io_channel_close(p->io);
+		g_io_channel_shutdown(p->io, TRUE, NULL);
 		g_io_channel_unref(p->io);
 	}
 
@@ -509,21 +538,21 @@ static gboolean ring_timer_cb(gpointer data)
 	return TRUE;
 }
 
-static void sco_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
-			const bdaddr_t *dst, gpointer user_data)
+static void sco_connect_cb(GIOChannel *chan, GError *err, gpointer user_data)
 {
 	int sk;
 	struct audio_device *dev = user_data;
 	struct headset *hs = dev->headset;
 	struct pending_connect *p = hs->pending;
 
-	if (err < 0) {
-		error("connect(): %s (%d)", strerror(-err), -err);
+	if (err) {
+		error("%s", err->message);
 
-		if (p->msg)
+		if (p && p->msg)
 			error_connection_attempt_failed(dev->conn, p->msg, p->err);
 
 		pending_connect_finalize(dev);
+
 		if (hs->rfcomm)
 			headset_set_state(dev, HEADSET_STATE_CONNECTED);
 		else
@@ -537,16 +566,19 @@ static void sco_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 	sk = g_io_channel_unix_get_fd(chan);
 
 	debug("SCO fd=%d", sk);
-	hs->sco = chan;
-	p->io = NULL;
+	hs->sco = g_io_channel_ref(chan);
 
-	if (p->msg) {
-		DBusMessage *reply = dbus_message_new_method_return(p->msg);
-		g_dbus_send_message(dev->conn, reply);
+	if (p) {
+		p->io = NULL;
+		if (p->msg) {
+			DBusMessage *reply;
+			reply = dbus_message_new_method_return(p->msg);
+			g_dbus_send_message(dev->conn, reply);
+		}
+
+		pending_connect_finalize(dev);
 	}
 
-	pending_connect_finalize(dev);
-
 	fcntl(sk, F_SETFL, 0);
 
 	headset_set_state(dev, HEADSET_STATE_PLAYING);
@@ -564,17 +596,24 @@ static int sco_connect(struct audio_device *dev, headset_stream_cb_t cb,
 			void *user_data, unsigned int *cb_id)
 {
 	struct headset *hs = dev->headset;
-	int err;
+	GError *err = NULL;
+	GIOChannel *io;
 
 	if (hs->state != HEADSET_STATE_CONNECTED)
 		return -EINVAL;
 
-	err = bt_sco_connect(&dev->src, &dev->dst, sco_connect_cb, dev);
-	if (err < 0) {
-		error("connect: %s (%d)", strerror(-err), -err);
-		return err;
+	io = bt_io_connect(BT_IO_SCO, sco_connect_cb, dev, NULL, &err,
+				BT_IO_OPT_SOURCE_BDADDR, &dev->src,
+				BT_IO_OPT_DEST_BDADDR, &dev->dst,
+				BT_IO_OPT_INVALID);
+	if (!io) {
+		error("%s", err->message);
+		g_error_free(err);
+		return -EIO;
 	}
 
+	g_io_channel_unref(io);
+
 	headset_set_state(dev, HEADSET_STATE_PLAY_IN_PROGRESS);
 
 	pending_connect_init(hs, HEADSET_STATE_PLAYING);
@@ -728,9 +767,15 @@ static int call_hold(struct audio_device *dev, const char *buf)
 	return 0;
 }
 
-static int button_press(struct audio_device *device, const char *buf)
+int telephony_key_press_rsp(void *telephony_device, cme_error_t err)
 {
-	struct headset *hs = device->headset;
+	return telephony_generic_rsp(telephony_device, err);
+}
+
+static int key_press(struct audio_device *device, const char *buf)
+{
+	if (strlen(buf) < 9)
+		return -EINVAL;
 
 	g_dbus_emit_signal(device->conn, device->path,
 			AUDIO_HEADSET_INTERFACE, "AnswerRequested",
@@ -741,7 +786,9 @@ static int button_press(struct audio_device *device, const char *buf)
 		ag.ring_timer = 0;
 	}
 
-	return headset_send(hs, "\r\nOK\r\n");
+	telephony_key_press_req(device, &buf[8]);
+
+	return 0;
 }
 
 int telephony_answer_call_rsp(void *telephony_device, cme_error_t err)
@@ -1077,7 +1124,7 @@ static struct event event_callbacks[] = {
 	{ "AT+CMER", event_reporting },
 	{ "AT+CHLD", call_hold },
 	{ "AT+CHUP", terminate_call },
-	{ "AT+CKPD", button_press },
+	{ "AT+CKPD", key_press },
 	{ "AT+CLIP", cli_notification },
 	{ "AT+BTRH", response_and_hold },
 	{ "AT+BLDN", last_dialed_number },
@@ -1112,7 +1159,7 @@ static void close_sco(struct audio_device *device)
 	if (hs->sco) {
 		g_source_remove(hs->sco_id);
 		hs->sco_id = 0;
-		g_io_channel_close(hs->sco);
+		g_io_channel_shutdown(hs->sco, TRUE, NULL);
 		g_io_channel_unref(hs->sco);
 		hs->sco = NULL;
 	}
@@ -1131,8 +1178,10 @@ static gboolean rfcomm_io_cb(GIOChannel *chan, GIOCondition cond,
 
 	hs = device->headset;
 
-	if (cond & (G_IO_ERR | G_IO_HUP))
+	if (cond & (G_IO_ERR | G_IO_HUP)) {
+		debug("ERR or HUP on RFCOMM socket");
 		goto failed;
+	}
 
 	if (g_io_channel_read(chan, (gchar *) buf, sizeof(buf) - 1,
 				&bytes_read) != G_IO_ERROR_NONE)
@@ -1213,22 +1262,37 @@ static gboolean sco_cb(GIOChannel *chan, GIOCondition cond,
 	return FALSE;
 }
 
-static void rfcomm_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
-			const bdaddr_t *dst, gpointer user_data)
+void headset_connect_cb(GIOChannel *chan, GError *err, gpointer user_data)
 {
 	struct audio_device *dev = user_data;
 	struct headset *hs = dev->headset;
 	struct pending_connect *p = hs->pending;
 	char hs_address[18];
 
-	if (err < 0) {
-		error("connect(): %s (%d)", strerror(-err), -err);
+	if (err) {
+		error("%s", err->message);
 		goto failed;
 	}
 
+	/* For HFP telephony isn't ready just disconnect */
+	if (hs->hfp_active && !ag.telephony_ready) {
+		error("Unable to accept HFP connection since the telephony "
+				"subsystem isn't initialized");
+		goto failed;
+	}
+
+	if (hs->tmp_rfcomm) {
+		hs->rfcomm = hs->tmp_rfcomm;
+		hs->tmp_rfcomm = NULL;
+	} else
+		hs->rfcomm = g_io_channel_ref(chan);
+
 	ba2str(&dev->dst, hs_address);
-	hs->rfcomm = chan;
-	p->io = NULL;
+
+	if (p)
+		p->io = NULL;
+	else
+		hs->auto_dc = FALSE;
 
 	if (server_is_enabled(&dev->src, HANDSFREE_SVCLASS_ID) &&
 			hs->hfp_handle != 0)
@@ -1247,14 +1311,14 @@ static void rfcomm_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 
 	headset_set_state(dev, HEADSET_STATE_CONNECTED);
 
-	if (p->target_state == HEADSET_STATE_PLAYING) {
+	if (p && p->target_state == HEADSET_STATE_PLAYING) {
 		p->err = sco_connect(dev, NULL, NULL, NULL);
 		if (p->err < 0)
 			goto failed;
 		return;
 	}
 
-	if (p->msg) {
+	if (p && p->msg) {
 		DBusMessage *reply = dbus_message_new_method_return(p->msg);
 		g_dbus_send_message(dev->conn, reply);
 	}
@@ -1264,7 +1328,7 @@ static void rfcomm_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 	return;
 
 failed:
-	if (p->msg)
+	if (p && p->msg)
 		error_connection_attempt_failed(dev->conn, p->msg, p->err);
 	pending_connect_finalize(dev);
 	if (hs->rfcomm)
@@ -1389,7 +1453,8 @@ static int rfcomm_connect(struct audio_device *dev, headset_stream_cb_t cb,
 {
 	struct headset *hs = dev->headset;
 	char address[18];
-	int err;
+	GError *err = NULL;
+	GIOChannel *io;
 
 	if (hs->rfcomm_ch < 0)
 		return get_records(dev, cb, user_data, cb_id);
@@ -1399,13 +1464,20 @@ static int rfcomm_connect(struct audio_device *dev, headset_stream_cb_t cb,
 	debug("%s: Connecting to %s channel %d", dev->path, address,
 		hs->rfcomm_ch);
 
-	err = bt_rfcomm_connect(&dev->src, &dev->dst, hs->rfcomm_ch,
-			rfcomm_connect_cb, dev);
-	if (err < 0) {
-		error("connect() failed: %s (%d)", strerror(-err), -err);
-		return err;
+	io = bt_io_connect(BT_IO_RFCOMM, headset_connect_cb, dev,
+				NULL, &err,
+				BT_IO_OPT_SOURCE_BDADDR, &dev->src,
+				BT_IO_OPT_DEST_BDADDR, &dev->dst,
+				BT_IO_OPT_CHANNEL, hs->rfcomm_ch,
+				BT_IO_OPT_INVALID);
+	if (!io) {
+		error("%s", err->message);
+		g_error_free(err);
+		return -EIO;
 	}
 
+	g_io_channel_unref(io);
+
 	headset_set_state(dev, HEADSET_STATE_CONNECT_IN_PROGRESS);
 
 	pending_connect_init(hs, HEADSET_STATE_CONNECTED);
@@ -1466,13 +1538,8 @@ static DBusMessage *hs_disconnect(DBusConnection *conn, DBusMessage *msg,
 {
 	struct audio_device *device = data;
 	struct headset *hs = device->headset;
-	DBusMessage *reply = NULL;
 	char hs_address[18];
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return NULL;
-
 	if (hs->state == HEADSET_STATE_DISCONNECTED)
 		return g_dbus_create_error(msg, ERROR_INTERFACE
 						".NotConnected",
@@ -1482,7 +1549,8 @@ static DBusMessage *hs_disconnect(DBusConnection *conn, DBusMessage *msg,
 	ba2str(&device->dst, hs_address);
 	info("Disconnected from %s, %s", hs_address, device->path);
 
-	return reply;
+	return dbus_message_new_method_return(msg);
+
 }
 
 static DBusMessage *hs_is_connected(DBusConnection *conn,
@@ -1528,6 +1596,8 @@ static DBusMessage *hs_connect(DBusConnection *conn, DBusMessage *msg,
 		return g_dbus_create_error(msg, ERROR_INTERFACE ".NotAllowed",
 						"Too many connected devices");
 
+	device->auto_connect = FALSE;
+
 	err = rfcomm_connect(device, NULL, NULL, NULL);
 	if (err < 0)
 		return g_dbus_create_error(msg, ERROR_INTERFACE
@@ -1789,6 +1859,7 @@ static DBusMessage *hs_get_properties(DBusConnection *conn,
 	DBusMessageIter iter;
 	DBusMessageIter dict;
 	gboolean value;
+	const char *state;
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
@@ -1806,6 +1877,11 @@ static DBusMessage *hs_get_properties(DBusConnection *conn,
 	value = (device->headset->state == HEADSET_STATE_PLAYING);
 	dict_append_entry(&dict, "Playing", DBUS_TYPE_BOOLEAN, &value);
 
+	/* State */
+	state = state2str(device->headset->state);
+	if (state)
+		dict_append_entry(&dict, "State", DBUS_TYPE_STRING, &state);
+
 	/* Connected */
 	value = (device->headset->state >= HEADSET_STATE_CONNECTED);
 	dict_append_entry(&dict, "Connected", DBUS_TYPE_BOOLEAN, &value);
@@ -1983,12 +2059,12 @@ static void headset_free(struct audio_device *dev)
 	}
 
 	if (hs->sco) {
-		g_io_channel_close(hs->sco);
+		g_io_channel_shutdown(hs->sco, TRUE, NULL);
 		g_io_channel_unref(hs->sco);
 	}
 
 	if (hs->rfcomm) {
-		g_io_channel_close(hs->rfcomm);
+		g_io_channel_shutdown(hs->rfcomm, TRUE, NULL);
 		g_io_channel_unref(hs->rfcomm);
 	}
 
@@ -2006,7 +2082,7 @@ static void path_unregister(void *data)
 		headset_set_state(dev, HEADSET_STATE_DISCONNECTED);
 	}
 
-	info("Unregistered interface %s on path %s",
+	debug("Unregistered interface %s on path %s",
 		AUDIO_HEADSET_INTERFACE, dev->path);
 
 	headset_free(dev);
@@ -2062,7 +2138,7 @@ register_iface:
 		return NULL;
 	}
 
-	info("Registered interface %s on path %s",
+	debug("Registered interface %s on path %s",
 		AUDIO_HEADSET_INTERFACE, dev->path);
 
 	return hs;
@@ -2149,7 +2225,6 @@ static gboolean dummy_connect_complete(struct audio_device *dev)
 
 unsigned int headset_request_stream(struct audio_device *dev,
 					headset_stream_cb_t cb,
-					headset_lock_t lock,
 					void *user_data)
 {
 	struct headset *hs = dev->headset;
@@ -2186,14 +2261,10 @@ unsigned int headset_request_stream(struct audio_device *dev,
 
 unsigned int headset_config_stream(struct audio_device *dev,
 					headset_stream_cb_t cb,
-					headset_lock_t lock,
 					void *user_data)
 {
 	struct headset *hs = dev->headset;
-	unsigned int id;
-
-	if (hs->lock & lock)
-		return 0;
+	unsigned int id = 0;
 
 	if (hs->dc_timer) {
 		g_source_remove(hs->dc_timer);
@@ -2223,24 +2294,21 @@ done:
 
 unsigned int headset_suspend_stream(struct audio_device *dev,
 					headset_stream_cb_t cb,
-					headset_lock_t lock,
 					void *user_data)
 {
 	struct headset *hs = dev->headset;
 	unsigned int id;
 
-	if (hs->lock & ~lock || !hs->rfcomm || !hs->sco)
-		return 0;
-
 	if (hs->dc_timer) {
 		g_source_remove(hs->dc_timer);
 		hs->dc_timer = 0;
 	}
 
-	close_sco(dev);
+	headset_set_state(dev, HEADSET_STATE_CONNECTED);
 
 	id = connect_cb_new(hs, HEADSET_STATE_CONNECTED, cb, user_data);
 	g_idle_add((GSourceFunc) dummy_connect_complete, dev);
+
 	return id;
 }
 
@@ -2258,13 +2326,23 @@ void set_hfp_active(struct audio_device *dev, gboolean active)
 	hs->hfp_active = active;
 }
 
+GIOChannel *headset_get_rfcomm(struct audio_device *dev)
+{
+	struct headset *hs = dev->headset;
+
+	return hs->tmp_rfcomm;
+}
+
 int headset_connect_rfcomm(struct audio_device *dev, GIOChannel *io)
 {
 	struct headset *hs = dev->headset;
 
-	hs->tmp_rfcomm = io;
+	if (hs->tmp_rfcomm)
+		return -EALREADY;
 
-	return hs->tmp_rfcomm ? 0 : -EINVAL;
+	hs->tmp_rfcomm = g_io_channel_ref(io);
+
+	return 0;
 }
 
 int headset_connect_sco(struct audio_device *dev, GIOChannel *io)
@@ -2274,7 +2352,7 @@ int headset_connect_sco(struct audio_device *dev, GIOChannel *io)
 	if (hs->sco)
 		return -EISCONN;
 
-	hs->sco = io;
+	hs->sco = g_io_channel_ref(io);
 
 	if (hs->pending_ring) {
 		ring_timer_cb(NULL);
@@ -2293,7 +2371,7 @@ static int headset_close_rfcomm(struct audio_device *dev)
 	GIOChannel *rfcomm = hs->tmp_rfcomm ? hs->tmp_rfcomm : hs->rfcomm;
 
 	if (rfcomm) {
-		g_io_channel_close(rfcomm);
+		g_io_channel_shutdown(rfcomm, TRUE, NULL);
 		g_io_channel_unref(rfcomm);
 		hs->tmp_rfcomm = NULL;
 		hs->rfcomm = NULL;
@@ -2307,59 +2385,50 @@ static int headset_close_rfcomm(struct audio_device *dev)
 	return 0;
 }
 
-void headset_set_authorized(struct audio_device *dev)
-{
-	struct headset *hs = dev->headset;
-
-	/* For HFP telephony isn't ready just disconnect */
-	if (hs->hfp_active && !ag.telephony_ready) {
-		error("Unable to accept HFP connection since the telephony "
-				"subsystem isn't initialized");
-		headset_set_state(dev, HEADSET_STATE_DISCONNECTED);
-		return;
-	}
-
-	hs->rfcomm = hs->tmp_rfcomm;
-	hs->tmp_rfcomm = NULL;
-
-	g_io_add_watch(hs->rfcomm,
-			G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-			(GIOFunc) rfcomm_io_cb, dev);
-
-	hs->auto_dc = FALSE;
-
-	/* For HSP (no special SLC setup) move to CONNECTED state */
-	if (!hs->hfp_active)
-		headset_set_state(dev, HEADSET_STATE_CONNECTED);
-}
-
 void headset_set_state(struct audio_device *dev, headset_state_t state)
 {
 	struct headset *hs = dev->headset;
 	gboolean value;
+	const char *state_str;
+	headset_state_t old_state = hs->state;
+	GSList *l;
 
-	if (hs->state == state)
+	if (old_state == state)
 		return;
 
+	state_str = state2str(state);
+
 	switch (state) {
 	case HEADSET_STATE_DISCONNECTED:
 		value = FALSE;
 		close_sco(dev);
 		headset_close_rfcomm(dev);
+		emit_property_changed(dev->conn, dev->path,
+					AUDIO_HEADSET_INTERFACE, "State",
+					DBUS_TYPE_STRING, &state_str);
 		g_dbus_emit_signal(dev->conn, dev->path,
 					AUDIO_HEADSET_INTERFACE,
 					"Disconnected",
 					DBUS_TYPE_INVALID);
-		emit_property_changed(dev->conn, dev->path,
+		if (hs->state > HEADSET_STATE_CONNECT_IN_PROGRESS)
+			emit_property_changed(dev->conn, dev->path,
 					AUDIO_HEADSET_INTERFACE, "Connected",
 					DBUS_TYPE_BOOLEAN, &value);
+
 		telephony_device_disconnected(dev);
 		active_devices = g_slist_remove(active_devices, dev);
 		break;
 	case HEADSET_STATE_CONNECT_IN_PROGRESS:
+		emit_property_changed(dev->conn, dev->path,
+					AUDIO_HEADSET_INTERFACE, "State",
+					DBUS_TYPE_STRING, &state_str);
 		break;
 	case HEADSET_STATE_CONNECTED:
 		close_sco(dev);
+		if (hs->state != HEADSET_STATE_PLAY_IN_PROGRESS)
+			emit_property_changed(dev->conn, dev->path,
+					AUDIO_HEADSET_INTERFACE, "State",
+					DBUS_TYPE_STRING, &state_str);
 		if (hs->state < state) {
 			value = TRUE;
 			g_dbus_emit_signal(dev->conn, dev->path,
@@ -2388,6 +2457,9 @@ void headset_set_state(struct audio_device *dev, headset_state_t state)
 		break;
 	case HEADSET_STATE_PLAYING:
 		value = TRUE;
+		emit_property_changed(dev->conn, dev->path,
+					AUDIO_HEADSET_INTERFACE, "State",
+					DBUS_TYPE_STRING, &state_str);
 		hs->sco_id = g_io_add_watch(hs->sco,
 					G_IO_ERR | G_IO_HUP | G_IO_NVAL,
 					(GIOFunc) sco_cb, dev);
@@ -2406,9 +2478,15 @@ void headset_set_state(struct audio_device *dev, headset_state_t state)
 		break;
 	}
 
-	debug("State changed %s: %s -> %s", dev->path, str_state[hs->state],
-		str_state[state]);
 	hs->state = state;
+
+	debug("State changed %s: %s -> %s", dev->path, str_state[old_state],
+		str_state[state]);
+
+	for (l = headset_callbacks; l != NULL; l = l->next) {
+		struct headset_state_callback *cb = l->data;
+		cb->cb(dev, old_state, state, cb->user_data);
+	}
 }
 
 headset_state_t headset_get_state(struct audio_device *dev)
@@ -2435,6 +2513,13 @@ gboolean headset_is_active(struct audio_device *dev)
 	return FALSE;
 }
 
+headset_lock_t headset_get_lock(struct audio_device *dev)
+{
+	struct headset *hs = dev->headset;
+
+	return hs->lock;
+}
+
 gboolean headset_lock(struct audio_device *dev, headset_lock_t lock)
 {
 	struct headset *hs = dev->headset;
@@ -2501,6 +2586,11 @@ gboolean headset_get_nrec(struct audio_device *dev)
 	return hs->nrec;
 }
 
+gboolean headset_get_sco_hci(struct audio_device *dev)
+{
+	return sco_hci;
+}
+
 int telephony_event_ind(int index)
 {
 	if (!active_devices)
@@ -2552,6 +2642,11 @@ int telephony_incoming_call_ind(const char *number, int type)
 		return -EBUSY;
 	}
 
+	/* With HSP 1.2 the RING messages should *not* be sent if inband
+	 * ringtone is being used */
+	if (!hs->hfp_active && (ag.features & AG_FEATURE_INBAND_RINGTONE))
+		return 0;
+
 	g_free(ag.number);
 	ag.number = g_strdup(number);
 	ag.number_type = type;
@@ -2671,3 +2766,34 @@ int telephony_call_waiting_ind(const char *number, int type)
 
 	return 0;
 }
+
+unsigned int headset_add_state_cb(headset_state_cb cb, void *user_data)
+{
+	struct headset_state_callback *state_cb;
+	static unsigned int id = 0;
+
+	state_cb = g_new(struct headset_state_callback, 1);
+	state_cb->cb = cb;
+	state_cb->user_data = user_data;
+	state_cb->id = ++id;
+
+	headset_callbacks = g_slist_append(headset_callbacks, state_cb);
+
+	return state_cb->id;
+}
+
+gboolean headset_remove_state_cb(unsigned int id)
+{
+	GSList *l;
+
+	for (l = headset_callbacks; l != NULL; l = l->next) {
+		struct headset_state_callback *cb = l->data;
+		if (cb && cb->id == id) {
+			headset_callbacks = g_slist_remove(headset_callbacks, cb);
+			g_free(cb);
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
